from abc import abstractmethod

from rltg.logic.PartialRewardAutomaton import PartialRewardAutomaton
from rltg.logic.RewardAutomaton import RewardAutomaton


class RewardShaping(object):

    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, state, state2, *args, is_terminal_state=False, **kwargs):
        return self.shaping_reward(state, state2, is_terminal_state=is_terminal_state)

    @abstractmethod
    def potential_function(self, s, is_terminal_state=False):
        raise NotImplementedError

    def shaping_reward(self, state, state2, is_terminal_state=False):
        phi = self.potential_function
        reward = phi(state2, is_terminal_state=is_terminal_state) - phi(state)
        # if state!=state2:
        #     print(state, state2, phi(state), phi(state2, is_terminal_state=is_terminal_state), reward)
        return reward


class AutomatonRewardShaping(RewardShaping):
    def __init__(self, automaton:RewardAutomaton, *args, **kwargs):
        super().__init__()
        self.automaton = automaton

    def potential_function(self, s, is_terminal_state=False):
        return self.automaton.potential_function(s, is_terminal_state=is_terminal_state)

class DynamicAutomatonRewardShaping(RewardShaping):
    def __init__(self, automaton:PartialRewardAutomaton, *args, **kwargs):
        super().__init__()
        self.automaton = automaton

    def potential_function(self, s, is_terminal_state=False):
        raise NotImplementedError

    def shaping_reward(self, state, state2, is_terminal_state=False):
        phi = self.automaton.potential_function
        reward = phi(state2, is_prime=True, is_terminal_state=is_terminal_state) - phi(state)
        return reward



import numpy as np

from rltg.agents.parameters.Parameter import AnnealedParameter
from rltg.agents.policies.Policy import Policy


class EGreedy(Policy):

    def __init__(self, epsilon=0.1, epsilon_start=None, decay_steps=10000, eval:bool=False):
        super().__init__(eval)
        if epsilon_start is None:
            epsilon_start = epsilon
        self.epsilon = AnnealedParameter(end=epsilon, start=epsilon_start, decay_steps=decay_steps)
        self.eval = eval


    def update(self, *args, **kwargs):
        self.epsilon.update()

    def choose_action(self, values=(), optimal=False, **kwargs):
        acts = np.arange(len(values))

        if not optimal and np.random.rand() < self.epsilon.get():
            action_idx = np.random.choice(acts)
        else:
            max_acts = acts[values == np.max(values)]
            # when multiple max actions, in training 'choice', in evaluation should be deterministic
            if self.eval:
                action_idx = max_acts[0]
            elif len(max_acts)==0:
                return 0
            else:
                action_idx = np.random.choice(max_acts)


        return action_idx



import gym
from gym import Wrapper
from gym.utils import EzPickle


class GoalEnvWrapper(Wrapper, EzPickle):
    """A OpenAI Gym environment goal-based."""

    def __init__(self, env, f):
        Wrapper.__init__(self, env)
        EzPickle.__init__(self, env, f)
        self.f = f

    def step(self, action):
        obs, reward, done, info = super().step(action)
        info["goal"] = self.f(obs, reward, done, info)
        return obs, reward, done, info



import numpy as np
from enum import Enum

from rltg.utils.misc import mydefaultdict


class Parameter(object):


    def update(self, *args, **kwargs):
        pass

    def get(self, *args, **kwargs):
        raise NotImplementedError


class Constant(Parameter):
    def __init__(self, c):
        self.c = c

    def get(self, *args, **kwargs):
        return self.c

class DecayType(Enum):
    LINEAR = 'linear'
    MULTIPLICATIVE = 'multiplicative'


class AnnealedParameter(Parameter):
    def __init__(self, start=0.0, end=0.0, decay_steps=10000, decay_factor=0.99, decay_type:DecayType=DecayType.LINEAR):
        """decay_steps applies only if decay_type=LINEAR
        decay_factor applies only if decay_type=MULTIPLICATIVE"""
        assert start>=end and decay_factor < 1.0
        self.start = start
        self.end = end

        self.decay_type = decay_type
        self.decay_steps = decay_steps
        self.decay_factor = decay_factor

        self.cur_value = self.start
        self.linear_decay_step = (start - end)/decay_steps

    def decay_function(self, x):
        if x <= self.end:
            return self.end
        if self.decay_type==DecayType.LINEAR:
            return  \
                x - self.linear_decay_step
        elif self.decay_type==DecayType.MULTIPLICATIVE:
            return self.end if x<=self.end else x*self.decay_factor


    def get(self, *args, **kwargs):
        return self.cur_value

    def update(self, *args, **kwargs):
        self.cur_value = self.decay_function(self.cur_value)


class AlphaVisitDecay(Parameter):
    def __init__(self, action_space):
        self.action_space = action_space
        self.Visits = mydefaultdict(np.zeros(self.action_space.n))

    def get(self, x, a, *args, **kwargs):
        return 1.0/(self.Visits[x][a])

    def update(self, x, a, *args, **kwargs):
        self.incVisits(x, a)

    def setVisits(self, x, a, q):
        self.Visits[x][a] = q

    def getSumVisits(self, x):
        return np.sum(self.Visits[x,:])

    def incVisits(self, x, a):
        self.Visits[x][a] += 1
